#include "stdafx.h"
#include "HTMLSyntaxColorer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

HTMLSyntaxColorState::HTMLSyntaxColorState()
{
	state = 0;
}

HTMLSyntaxColorState::HTMLSyntaxColorState(HTMLSyntaxColorState* original)
{
	stateStack = original->stateStack;
	state = original->state;
}

/************* Scanning a HTML file ******************/

#define T_ILL    0
#define T_HTML   1
#define T_HEAD   2
#define T_TITLE  3
#define T_BODY   4
#define T_ADDR   5
#define T_LINK   6
#define T_H     10
#define T_VERB  11
#define T_DIR   12
#define T_LIST  13
#define T_DESC  14
#define T_ITEM  15
#define T_DT    16
#define T_DD    17
#define T_P     18
#define T_A     19
#define T_IMG   20
#define T_CHAR  21
#define T_BR    22
#define T_META  23
#define T_SCRIPT 24
#define T_STRUCT 25
#define NR_TAGS   80


#define TN_H1   5

#define C_NO    0
#define C_YES   1
#define C_OPT   2


struct Codes
  { char *name;
    int  closing;
    byte kind;
  } tags[NR_TAGS] =
{
#define H_HTML     0
   { "html",    C_YES, T_HTML },
#define H_HEAD     1
   { "head",    C_YES, T_HEAD },
#define H_TITLE    2
   { "title",   C_YES, T_TITLE },
#define H_BODY     3
   { "body",    C_YES, T_BODY },
#define H_ADDRESS  4
   { "address", C_YES, T_ADDR },
#define H_H1       5
   { "h1",      C_YES, T_H },
#define H_H2       6
   { "h2",      C_YES, T_H },
#define H_H3       7
   { "h3",      C_YES, T_H },
#define H_H4       8
   { "h4",      C_YES, T_H },
#define H_H5       9
   { "h5",      C_YES, T_H },
#define H_H6      10
   { "h6",      C_YES, T_H },
#define H_P       11
   { "p",       C_OPT, T_P },
#define H_UL      12
   { "ul",      C_YES, T_LIST },
#define H_MENU    13
   { "menu",    C_YES, T_LIST },
#define H_DIR     14
   { "dir",     C_YES, T_LIST },
#define H_OL      15
   { "ol",      C_YES, T_LIST },
#define H_LI      16
   { "li",      C_OPT, T_ITEM },
#define H_LH      17
   { "lh",      C_OPT, T_ITEM },
#define H_DL      18
   { "dl",      C_YES, T_DESC },
#define H_DT      19
   { "dt",      C_OPT, T_DT },
#define H_DD      20
   { "dd",      C_OPT, T_DD },
#define H_A       21
   { "a",       C_OPT, T_A },
#define H_Q       22
   { "q",       C_YES, T_CHAR },
#define H_I       23
   { "i",       C_YES, T_CHAR },
#define H_EM      24
   { "em",      C_YES, T_CHAR },
#define H_B       25
   { "b",       C_YES, T_CHAR },
#define H_STRONG  26
   { "strong",  C_YES, T_CHAR },
#define H_TT      27
   { "tt",      C_YES, T_CHAR },
#define H_SAMP    28
   { "samp",    C_YES, T_CHAR },
#define H_KDB     29
   { "kbd",     C_YES, T_CHAR },
#define H_VAR     30
   { "var",     C_YES, T_CHAR },
#define H_DFN     31
   { "dfn",     C_YES, T_CHAR },
#define H_CODE    32
   { "code",    C_YES, T_CHAR },
#define H_BLINK   33
   { "blink",   C_YES, T_CHAR },
#define H_CITE    34
   { "cite",    C_YES, T_CHAR },
#define H_BLOCKQUOTE 35
   { "blockquote", C_YES, T_CHAR },
#define H_BQ      36
   { "bq",      C_YES, T_CHAR },
#define H_U       37
   { "u",       C_YES, T_CHAR },
#define H_S       38
   { "s",       C_YES, T_CHAR },
#define H_SMALL   39
   { "small",   C_YES, T_CHAR },
#define H_BIG     40
   { "big",     C_YES, T_CHAR },
#define H_NOTE    41
   { "note",    C_YES, T_CHAR },
#define H_AU      42
   { "au",      C_YES, T_CHAR },
#define H_PERSON  43
   { "person",  C_YES, T_CHAR },
#define H_ACRONYM 44
   { "acronym", C_YES, T_CHAR },
#define H_ABBREV  45
   { "abbrev",  C_YES, T_CHAR },
#define H_CREDIT  46
   { "credit",  C_YES, T_CHAR },
#define H_INS     47
   { "ins",     C_YES, T_CHAR },
#define H_DEL     48
   { "del",     C_YES, T_CHAR },
#define H_PRE     49
   { "pre",     C_YES, T_VERB },
#define H_XMP     50
   { "xmp",     C_YES, T_VERB },
#define H_LISTING 51
   { "listing", C_YES, T_VERB },
#define H_BR      52
   { "br",      C_NO,  T_DIR },
#define H_HR      53
   { "hr",      C_NO,  T_DIR },
#define H_IMG     54
   { "img",     C_NO,  T_IMG },
#define H_ISINDEX 55
   { "isindex", C_NO,  T_DIR },
#define H_SELECT  56
   { "select",  C_YES, T_DIR },
#define H_LINK    57
   { "link",    C_NO,  T_LINK },
#define H_CENTER  58
   { "center",  C_YES, T_CHAR },
#define H_META    59
   { "meta",    C_NO,  T_META },
#define H_TABLE   60
   { "table",   C_YES, T_DIR },
#define H_TH      61
   { "th",      C_OPT, T_DIR },
#define H_TR      62
   { "tr",      C_OPT, T_DIR },
#define H_TD      63
   { "td",      C_OPT, T_DIR },
#define H_SUP     64
   { "sup",     C_YES, T_DIR },
#define H_SUB     65
   { "sub",     C_YES, T_DIR },
#define H_CAPTION 66
   { "caption", C_OPT, T_CHAR },
#define H_SCRIPT  67
   { "script",  C_OPT, T_SCRIPT },
#define H_FONT    68
   { "font",    C_OPT, T_DIR },
#define H_FORM    69
   { "form",    C_OPT, T_CHAR },
#define H_INPUT   70
   { "input",   C_NO,  T_CHAR },
#define H_TEXTAREA 71
   { "textarea",C_OPT, T_CHAR },   
#define H_NOBR      72
   { "nobr",    C_OPT,  T_DIR },
#define H_NOSCRIPT  73
   { "noscript", C_YES, T_DIR },
#define H_MARQUEE   74
   { "marquee", C_YES, T_DIR },
#define H_STYLE     75
   { "style", C_YES, T_DIR },
#define H_IFORM     76
   { "iform", C_YES, T_STRUCT },
#define H_OBJECT    77 
   { "object", C_YES, T_STRUCT },
#define H_PARAM     78
   { "param", C_NO, T_STRUCT },
#define H_EMBED     79
   { "embed", C_YES, T_STRUCT },
};

#define H_ROUND_BRACKER  100
#define H_SQAURE_BRACKER 101
#define H_CURLY_BRACKER  102

int find_tag(const char *name)
{
	for (int i = 0; i < NR_TAGS; i++)
		if (strcmp(name, tags[i].name) == 0)
			return i;
	return -1;
} 

#define AT_ANY   0
#define AT_ALIGN 1
#define AT_COLOR 2
#define AT_NUM   3
#define AT_IMG_ALIGN 4
#define AT_FILE  5

typedef struct {
  int tagnr;
  char *attr_name;
  int type;	  
} valid_comb_t;
valid_comb_t valid_comb[] = {
  { H_BODY,  "alink", AT_COLOR },
  { H_BODY,  "onload", AT_ANY },
  { H_BODY,  "bgcolor", AT_COLOR },
  { H_BODY,  "bgproperties", AT_ANY },
  { H_BODY,  "background", AT_FILE },
  { H_P,     "title", AT_ANY },
  { H_P,     "align", AT_ALIGN },
  { H_TABLE, "width", AT_NUM },
  { H_TABLE, "border", AT_NUM },
  { H_TABLE, "cellpadding", AT_NUM },
  { H_TABLE, "cellspacing", AT_NUM },
  { H_TABLE, "align", AT_ALIGN },
  { H_TABLE, "vspace", AT_NUM },
  { H_TABLE, "hspace", AT_NUM },
  { H_TABLE, "bgcolor", AT_COLOR },
  { H_TABLE, "background", AT_FILE },
  { H_TR,    "valign", AT_ANY },
  { H_TR,    "align", AT_ALIGN },
  { H_IMG,   "border", AT_NUM },
  { H_IMG,   "width", AT_NUM },
  { H_IMG,   "height", AT_NUM },
  { H_IMG,   "align", AT_IMG_ALIGN },
  { H_IMG,   "space", AT_NUM },
  { H_IMG,   "hspace", AT_NUM },
  { H_IMG,   "vspace", AT_NUM },
  { H_IMG,   "alt", AT_ANY },
  { H_IMG,   "nosave", AT_ANY },
  { H_IMG,   "src", AT_FILE },
  { H_FONT,  "color", AT_COLOR },
  { H_FONT,  "size", AT_NUM },
  { H_FONT,  "face", AT_ANY },
  { H_BR,    "clear", AT_ANY },
  { H_HR,    "width", AT_NUM },
  { H_HR,    "align", AT_ALIGN },
  { H_H1,    "align", AT_ALIGN },
  { H_H2,    "align", AT_ALIGN },
  { H_H3,    "align", AT_ALIGN },
  { H_H4,    "align", AT_ALIGN },
  { H_OL,    "type", AT_NUM },
  { H_UL,    "type", AT_NUM },
  { H_FORM,  "method", AT_ANY },
  { H_FORM,  "action", AT_ANY },
  { H_FORM,  "name", AT_ANY },
  { H_FORM,  "target", AT_ANY },
  { H_INPUT, "type", AT_ANY },
  { H_INPUT, "name", AT_ANY },
  { H_INPUT, "size", AT_NUM },
  { H_INPUT, "maxlength", AT_NUM },
  { H_INPUT, "value", AT_ANY },
  { H_INPUT, "onclick", AT_ANY },
  { H_INPUT, "checked", AT_ANY },
  { H_TEXTAREA, "name", AT_ANY },
  { H_TEXTAREA, "cols", AT_NUM },
  { H_TEXTAREA, "rows", AT_NUM },
  { H_TD,    "align", AT_ALIGN },
  { H_TD,    "valign", AT_ANY },
  { H_TD,    "colspan", AT_ANY },
  { H_TD,    "rowspan", AT_ANY },
  { H_TD,    "width", AT_NUM },
  { H_TD,    "bgcolor", AT_COLOR },
  { H_TD,    "style", AT_ANY },
  { H_TD,    "title", AT_ANY },
  { H_TH,    "align", AT_ALIGN },
  { H_TH,    "valign", AT_ANY },
  { H_TH,    "colspan", AT_ANY },
  { H_TH,    "rowspan", AT_ANY },
  { H_TH,    "width", AT_NUM },
  { H_TH,    "bgcolor", AT_COLOR },
  { H_TH,    "style", AT_ANY },
  { H_TH,    "title", AT_ANY },
  { H_A,     "name", AT_ANY },
  { H_A,     "href", AT_ANY },
  { H_A,     "target", AT_ANY },
  { H_A,     "onmouseover", AT_ANY },
  { H_A,     "onclick", AT_ANY },
  { H_A,     "title", AT_ANY },
  { H_A,     "style", AT_ANY },
  { H_SCRIPT, "language", AT_ANY },
  { H_SCRIPT, "src", AT_FILE },
  { H_MARQUEE, "direction", AT_ANY },
  { H_MARQUEE, "height", AT_ANY },
  { H_MARQUEE, "scrollamount", AT_ANY },
  { H_STYLE,  "type", AT_ANY },
  { H_META,	  "name", AT_ANY },
  { H_META,	  "content", AT_ANY },
  { H_OBJECT, "classid", AT_ANY },
  { H_OBJECT, "codebase", AT_ANY },
  { H_OBJECT, "width", AT_ANY },
  { H_OBJECT, "height", AT_ANY },
  { H_PARAM, "name", AT_ANY },
  { H_PARAM, "value", AT_ANY },
  { H_EMBED, "src", AT_ANY },
  { H_EMBED, "quality", AT_ANY },
  { H_EMBED, "bgcolor", AT_ANY },
  { H_EMBED, "width", AT_ANY },
  { H_EMBED, "height", AT_ANY },
  { H_EMBED, "name", AT_ANY },
  { H_EMBED, "type", AT_ANY },
  { H_EMBED, "pluginspage", AT_ANY },
};

char *ch_table[] = {
/*160*/ "nbsp", "iexp", "cent", "pound", "curren", "yen", "brvar", "sect", "uml", "copy",
/*170*/ "ordf", "laquo", "not", "shy", "reg", "macr", "deg", "plusmin", "sup2", "sup3",
/*180*/ "acute", "micro", "para", "middot", "cedol", "sup1", "ordm", "raquo", "frac14", "fraq12",
/*190*/ "fraq34", "iquest", "Agrave", "Aacute", "Acirc", "Atilde", "Auml", "Aring", "AElig", "Ccedil",
/*200*/ "Egrave", "Eacute", "Ecirc", "Euml", "Igrave", "Iacute", "Icirc", "Iuml", "ETH", "Ntilde",
/*210*/ "Ograve", "Oacute", "Ocirc", "Otilde", "Ouml", "Oslash", "Ugrave", "Uacute", "Ucirc",
/*220*/ "Uuml", "Yacute", "THORN", "szlig", "agrave", "aacute", "acirc", "atilde", "auml", "aring",
/*230*/ "aelig", "ccedil", "egrave", "eacute", "ecirc", "euml", "igrave", "iacute", "icirc", "iuml",
/*240*/ "eth", "ntilde", "ograve", "oacute", "ocirc", "otilde", "ouml", "oslash", "ugrave",
/*250*/ "uacute", "ucirc", "uuml", "yacute", "thorn", "yuml", "aring", "Eth", "icirc", "Thorn",
        "Yuml", "nbsp", "emsp", "ensp", "shy", "pd", "emdash", "endash", "copy", "reg",
        "trade", "alpha", "beta", "gamma", "delta", "epsi", "zeta", "eta", "theta", "thetav",
        "iota", "kappa", "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma",
        "tau", "upsi", "phi", "chi", "psi", "omega", "Alpha", "Beta", "Gamma", "Delta",
        "Epsi", "Zeta", "Eta", "Theta", "Iota", "Kappa", "Lambda", "Mu", "Nu", "Xi",
        "Pi", "Rho", "Sigma", "Tau", "Upsi", "Phi", "Chi", "Psi", "Omega", "amp",
        "gt", "lt", "quot", "euro",
};

#define NR_CH_TABLE (sizeof(ch_table)/sizeof(char *))


#define ST_TEXT				0
#define ST_IN_TAG_OPEN		1	// between < and tagname
#define ST_IN_TAG			2	// after tagname or attribute
#define ST_IN_TAG_ATTR		3	// after attribute name
#define ST_IN_TAG_IS		4	// after =
#define ST_IN_TAG_CLOSE		6	// after /
#define ST_IN_TAG_META		7	// meta tag starting with !
#define ST_IN_TAG_COMM		8	// inside comment
#define ST_IN_TAG_COMM2		9	// inside comment (meta tag) -- the usual case

#define ST_IN_JS           10
#define ST_JS_COMMENT      11
#define ST_JS_STRING       12
#define ST_JS_CHAR         13


#define COLOR_WHITESPACE	RGB(255, 255, 255)
#define COLOR_NORMALTEXT	RGB(  0,   0,   0)
#define COLOR_COMMENT		RGB( 64,  64, 220)
#define COLOR_TEXT_ERROR	RGB(255,   0,   0)
#define COLOR_BK_ERROR		RGB(255, 240, 240)
#define COLOR_TEXT_TAG		RGB( 64,   0,  64)
#define COLOR_BK_TAG		RGB(255, 240, 255)
#define COLOR_TEXT_VAL		RGB(  0, 128,   0)
#define COLOR_BK_VAL		RGB(230, 255, 230)

#define COLOR_JSCOMMENT		RGB( 64,  64, 220)
#define COLOR_NUMBER		RGB(128,   0,   0)
#define COLOR_OPERATOR		RGB(  0,   0,   0)
#define COLOR_STRING		RGB(  0, 128,   0)
#define COLOR_BK_STRING		RGB(240, 255, 240)
#define COLOR_KEYWORD		RGB(  0,  64, 128)
#define COLOR_SELBKGND		RGB(  0,   0,   0)
#define COLOR_SELTEXT		RGB(255, 255, 255)

//	Java keywords
static LPTSTR s_apszJavaScriptKeywordList[] =
{
	_T("abstract"),
	_T("boolean"),
	_T("break"),
	_T("byte"),
	_T("case"),
	_T("catch"),
	_T("char"),
	_T("class"),
	_T("const"),
	_T("continue"),
	_T("debugger"),
	_T("default"),
	_T("delete"),
	_T("do"),
	_T("double"),
	_T("else"),
	_T("enum"),
	_T("export"),
	_T("extends"),
	_T("false"),
	_T("final"),
	_T("finally"),
	_T("float"),
	_T("for"),
	_T("function"),
	_T("goto"),
	_T("if"),
	_T("implements"),
	_T("import"),
	_T("in"),
	_T("instanceof"),
	_T("int"),
	_T("interface"),
	_T("long"),
	_T("native"),
	_T("new"),
	_T("null"),
	_T("package"),
	_T("private"),
	_T("protected"),
	_T("public"),
	_T("return"),
	_T("short"),
	_T("static"),
	_T("super"),
	_T("switch"),
	_T("synchronized"),
	_T("this"),
	_T("throw"),
	_T("throws"),
	_T("transient"),
	_T("true"),
	_T("try"),
	_T("typeof"),
	_T("var"),
	_T("void"),
	_T("volatile"),
	_T("while"),
	_T("with"),
	NULL
};

static BOOL IsJavaScriptKeyword(LPCTSTR pszChars, int nLength)
{
	for (int L = 0; s_apszJavaScriptKeywordList[L] != NULL; L ++)
	{
		if (strncmp(s_apszJavaScriptKeywordList[L], pszChars, nLength) == 0
				&& s_apszJavaScriptKeywordList[L][nLength] == 0)
			return TRUE;
	}
	return FALSE;
}

static BOOL IsJavaScriptNumber(LPCTSTR pszChars, int nLength)
{
	if (nLength > 2 && pszChars[0] == '0' && pszChars[1] == 'x')
	{
		for (int I = 2; I < nLength; I++)
		{
			if (isdigit(pszChars[I]) || (pszChars[I] >= 'A' && pszChars[I] <= 'F') ||
										(pszChars[I] >= 'a' && pszChars[I] <= 'f'))
				continue;
			return FALSE;
		}
		return TRUE;
	}
	if (! isdigit(pszChars[0]))
		return FALSE;
	for (int I = 1; I < nLength; I++)
	{
		if (! isdigit(pszChars[I]) && pszChars[I] != '+' &&
			pszChars[I] != '-' && pszChars[I] != '.' && pszChars[I] != 'e' &&
			pszChars[I] != 'E')
			return FALSE;
	}
	return TRUE;
}


HTMLSyntaxColorer::HTMLSyntaxColorer()
{
	defaultTextColor = ::GetSysColor(COLOR_WINDOWTEXT);
	defaultBkColor = ::GetSysColor(COLOR_WINDOW);
	noTextColor = RGB(240, 240, 240);
	selectTextColor = RGB(255, 255, 255);
	selectBkColor = RGB(0, 0, 0);
}

void HTMLSyntaxColorer::parseLine(const char* line, HTMLSyntaxColorState &HTMLstate, int &lineFlags)
{
	SharedStack<short> *stateStack = &HTMLstate.stateStack;
	int state = HTMLstate.state;
	BOOL bDecIndex  = FALSE;
	int nIdentBegin = -1;

	switch(state)
	{
	case ST_JS_COMMENT:
		startColor(0, COLOR_JSCOMMENT, COLOR_WHITESPACE);
		break;
	}

	int i = 0;
	while (line[i] != '\0')
	{
		switch(state)
		{
		case ST_TEXT:
			if (line[i] == '<')
			{	startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				state = ST_IN_TAG_OPEN;
				i++;
			}
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
				endColor(i);
			}
			else if (line[i] == '&')
			{	int begin = i;
				bool error = false;
				i++;
				if (line[i] == '#')
				{
					i++;
					long chnr = 0;
					for (; isdigit(line[i]); i++)
						chnr = chnr * 10 + line[i] - '0';
					if (   chnr >= 160 && chnr <= 255
						&& ch_table[chnr - 160] != 0)
						color(begin+2, i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				}
				else
				{
					char chcode[30];
					int j = 0;
					while (   line[i] != ';' && line[i] != '\0' 
						   && line[i] != '<' && line[i] != '>' && line[i] != '&')
					{	if (j < 29)
							chcode[j++] = line[i];
						i++;
					}
					chcode[j] = 0;
					for (j = 0; j < NR_CH_TABLE;j++)
						if (ch_table[j] != 0 && strcmp(ch_table[j], chcode) == 0)
							break;
					if (j == NR_CH_TABLE)
						error = true;

				}
				if (line[i] == ';')
					i++;
				else
					error = true;
				if (error)
					color(begin, i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
			}
			else
				i++;
			break;
		case ST_IN_TAG_OPEN:
			startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (line[i] == '!')
			{	state = ST_IN_TAG_META;
				i++;
			}
			else if (isalpha(line[i]))
			{	int begin = i;
				char tagname[30];
				int j = 0;
				while (isalnum(line[i]))
				{	if (j < 29)
						tagname[j++] = tolower(line[i]);
					i++;
				}
				tagname[j] = '\0';
				int tagnr = find_tag(tagname);
				if (tagnr < 0)
					startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				else
					stateStack->push((short)tagnr);
				state = ST_IN_TAG;
			}
			else if (line[i] == '/')
			{	int begin = i;
				i++;
				char tagname[30];
				int j = 0;
				while (isalnum(line[i]))
				{	if (j < 29)
						tagname[j++] = tolower(line[i]);
					i++;
				}
				tagname[j] = '\0';
				int tagnr = find_tag(tagname);
				if (   tagnr < 0 
					|| tags[tagnr].closing == C_NO
					|| stateStack->isEmpty())
					startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				else
				{
					// Close optional tags
					while (   !stateStack->isEmpty() 
						   && stateStack->top() != tagnr
						   && tags[stateStack->top()].closing == C_OPT)
						stateStack->pop();

					// Match?
					if (stateStack->topIs(tagnr))
						stateStack->pop();
					else
						startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				}
				state = ST_IN_TAG_CLOSE;
			}
			else
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			break;
		case ST_IN_TAG:
			startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (line[i] == '-' && line[i+1] == '-')
			{	state = ST_IN_TAG_COMM;
				startColor(i, COLOR_COMMENT, defaultBkColor);
				i += 2;
			}
			else if (isalpha(line[i]))
			{	int begin = i;
				char attrname[30];
				int j = 0;
				while (isalnum(line[i]))
				{	if (j < 29)
						attrname[j++] = tolower(line[i]);
					i++;
				}
				attrname[j] = '\0';
				if (stateStack->isEmpty())
					startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				else
				{	int tagnr = stateStack->top();

					for (int k = 0; k < sizeof(valid_comb)/sizeof(valid_comb[0]); k++)
						if (   valid_comb[k].tagnr == tagnr
							&& strcmp(valid_comb[k].attr_name, attrname) == 0)
							break;
					if (k == sizeof(valid_comb)/sizeof(valid_comb[0]))
						startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				}
				state = ST_IN_TAG_ATTR;
			}
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				if (!stateStack->isEmpty() && tags[stateStack->top()].closing == C_NO)
					stateStack->pop();
				state = ST_TEXT;
				i++;
				endColor(i);
			}
			else if (line[i] == '/')
			{	i++;
				state = ST_IN_TAG_CLOSE;
			}
			else
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			break;
		case ST_IN_TAG_ATTR:
			startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (isalpha(line[i]))
			{	state = ST_IN_TAG;
			}
			else if (line[i] == '=')
			{	state = ST_IN_TAG_IS;
				i++;
			}
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				if (!stateStack->isEmpty() && tags[stateStack->top()].closing == C_NO)
					stateStack->pop();
				state = ST_TEXT;
				i++;
				endColor(i);
			}
			else if (line[i] == '/')
			{	i++;
				state = ST_IN_TAG_CLOSE;
			}
			else
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			break;
		case ST_IN_TAG_IS:
			startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (line[i] == '"')
			{
				int begin = i;
				i++;
				while (line[i] != '\0' && line[i] != '"')
					i++;
				if (line[i] == '"')
				{	startColor(begin, COLOR_TEXT_VAL, COLOR_BK_TAG);
					i++;
					startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				}
				else
					startColor(begin, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				state = ST_IN_TAG;
			}
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
				endColor(i);
			}
			else
			{	
				startColor(i, COLOR_TEXT_VAL, COLOR_BK_TAG);
				while (line[i] != '\0' && line[i] != ' ' && line[i] != '>')
					i++;
				startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				state = ST_IN_TAG;
			}
			break;
		case ST_IN_TAG_COMM:
		case ST_IN_TAG_COMM2:
			startColor(i, COLOR_COMMENT, defaultBkColor);
			while (line[i] != '\0')
			{	if (line[i] == '-' && line[i+1] == '-')
				{	state = state == ST_IN_TAG_COMM ? ST_IN_TAG : ST_IN_TAG_META;
					i += 2;
					endColor(i);
					break;
				}
				i++;
			}
			break;
		case ST_IN_TAG_CLOSE:
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				if (!stateStack->isEmpty() && tags[stateStack->top()].closing == C_NO)
					stateStack->pop();
				state = ST_TEXT;
				i++;
				endColor(i);
			}
			else
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
				state = ST_IN_TAG;
			}
			break;
		case ST_IN_TAG_META:
			if (line[i] == ' ' || line[i] == '\t')
				i++;
			else if (line[i] == '-' && line[i+1] == '-')
			{	startColor(i, COLOR_COMMENT, defaultBkColor);
				i += 2;
				if (stateStack->topIs(H_SCRIPT))
					state = ST_IN_JS;
				else
					state = ST_IN_TAG_COMM2;
				
			}
			else if (isalpha(line[i]))
			{	while (line[i] != '\0' && line[i] != ' ' && line[i] != '\t'
					   && line[i] != '>')
					i++;
			}
			else if (line[i] == '"')
			{
				int begin = i;
				i++;
				while (line[i] != '\0' && line[i] != '"')
					i++;
				if (line[i] == '"')
				{	startColor(begin, COLOR_TEXT_VAL, COLOR_BK_TAG);
					i++;
					startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				}
			}
			else if (line[i] == '>')
			{	startColor(i, COLOR_TEXT_TAG, COLOR_BK_TAG);
				state = ST_TEXT;
				i++;
				endColor(i);
			}
			
			else
			{	startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			break;
		case ST_IN_JS:
			if (nIdentBegin >= 0 && !(isalnum(line[i]) || line[i] == '_' || line[i] == '.'))
			{
				if (IsJavaScriptKeyword(line + nIdentBegin, i - nIdentBegin))
				{
					startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
				}
				else if (IsJavaScriptNumber(line + nIdentBegin, i - nIdentBegin))
				{
					startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
				}
				endColor(i);
				nIdentBegin = -1;
			}
			if (line[i] == '-' && line[i+1] == '-')
			{
				bool error = false;
				while (stateStack->topIs(H_ROUND_BRACKER) || stateStack->topIs(H_SQAURE_BRACKER) || stateStack->topIs(H_CURLY_BRACKER))
				{
					error = true;
					stateStack->pop();
				}
				if (error)
					startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				else
					startColor(i, COLOR_COMMENT, defaultBkColor);
				i += 2;
				state = ST_IN_TAG_META;
			}
			else if (line[i] == '/' && line[i+1] == '/')
			{
				startColor(i, COLOR_JSCOMMENT, COLOR_WHITESPACE);
				i += 2;
				while (line[i] != '\0')
				{
					if (line[i] == '-' && line[i+1] == '-')
					{
						bool error = false;
						while (stateStack->topIs(H_ROUND_BRACKER) || stateStack->topIs(H_SQAURE_BRACKER) || stateStack->topIs(H_CURLY_BRACKER))
						{
							error = true;
							stateStack->pop();
						}
						if (error)
							startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
						else
							startColor(i, COLOR_COMMENT, defaultBkColor);
						i += 2;
						state = ST_IN_TAG_META;
						break;
					}
					i++;
				}
			}
			else if (line[i] == '/' && line[i+1] == '*')
			{
				startColor(i, COLOR_JSCOMMENT, COLOR_WHITESPACE);
				i += 2;
				state = ST_JS_COMMENT;
			}
			else if (line[i] == '"')
			{
				startColor(i+1, COLOR_STRING, COLOR_BK_STRING);
				state = ST_JS_STRING;
				i++;
			}
			else if (line[i] == '\'')
			{
				startColor(i+1, COLOR_STRING, COLOR_BK_STRING);
				state = ST_JS_CHAR;
				i++;
			}
			else if (line[i] == '(')
			{	
				stateStack->push(H_ROUND_BRACKER);
				i++;
			}
			else if (line[i] == ')')
			{	
				if (stateStack->topIs(H_ROUND_BRACKER))
					stateStack->pop();
				else
					startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			else if (line[i] == '[')
			{	
				stateStack->push(H_SQAURE_BRACKER);
				i++;
			}
			else if (line[i] == ']')
			{	if (stateStack->topIs(H_SQAURE_BRACKER))
					stateStack->pop();
				else
					startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			else if (line[i] == '{')
			{	
				if (stateStack->topIs(H_ROUND_BRACKER))
					startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				stateStack->push(H_CURLY_BRACKER);
				i++;
			}
			else if (line[i] == '}')
			{	
				if (stateStack->topIs(H_CURLY_BRACKER))
					stateStack->pop();
				else
					startColor(i, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				i++;
			}
			else if (isalnum(line[i]) || line[i] == '_' || line[i] == '.')
			{
				if (nIdentBegin == -1)
					nIdentBegin = i;
				i++;
			}
			else
			{
				i++;
			}
			break;
		case ST_JS_COMMENT:
			if (line[i] == '*' && line[i+1] == '/')
			{
				i += 2;
				state = ST_IN_JS;
				endColor(i);
			}
			else
				i++;
			break;
		case ST_JS_STRING:
			if (line[i] == '\\' && line[i] != '\0')
				i++;
			else if (line[i] == '"')
			{
				state = ST_IN_JS;
				endColor(i);
			}
			i++;
			break;
		case ST_JS_CHAR:
			if (line[i] == '\\' && line[i] != '\0')
				i++;
			else if (line[i] == '\'')
			{
				state = ST_IN_JS;
				endColor(i);
			}
			i++;
			break;
		}
	}

	if (nIdentBegin >= 0)
	{
		if (IsJavaScriptKeyword(line + nIdentBegin, i - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
		}
		else
		if (IsJavaScriptNumber(line + nIdentBegin, i - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
		}
	}

	HTMLstate.state = state;
}
