// SampleMultiDocTemplate.cpp : implementation of the CSampleMultiDocTemplate class
//

#include "stdafx.h"
#include "SampleMultiDocTemplate.h"
#include "Sample.h"
#include "gfl.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CSampleMultiDocTemplate

//IMPLEMENT_DYNCREATE(CSampleMultiDocTemplate, CMultiDocTemplate)

//BEGIN_MESSAGE_MAP(CSampleMultiDocTemplate, CMultiDocTemplate)
	//{{AFX_MSG_MAP(CSampleMultiDocTemplate)
	//}}AFX_MSG_MAP
//END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CSampleMultiDocTemplate construction/destruction

#pragma warning(disable:4355)
CSampleMultiDocTemplate::CSampleMultiDocTemplate(UINT nIDResource, CRuntimeClass* pDocClass, CRuntimeClass* pFrameClass, CRuntimeClass* pViewClass,
												 CRuntimeClass* pDocImageClass, CRuntimeClass* pViewImageClass)
		: CMultiDocTemplate(nIDResource, pDocClass, pFrameClass, pViewClass),
		  m_pDocImageClass(pDocImageClass),
		  m_pViewImageClass(pViewImageClass)
{
}

CSampleMultiDocTemplate::~CSampleMultiDocTemplate()
{
}

/////////////////////////////////////////////////////////////////////////////
// CSampleMultiDocTemplate commands

CDocument* CSampleMultiDocTemplate::OpenDocumentFile(LPCTSTR lpszPathName,
	BOOL bMakeVisible)
{
	// Close initial document, if it has not been modified yet
    if (m_docList.GetCount() == 1)
    {
        CDocument* pDocument = (CDocument*)m_docList.GetHead();
        if (pDocument->GetPathName().IsEmpty() && !pDocument->IsModified())
        {
			POSITION pos = pDocument->GetFirstViewPosition();
			if (pos != NULL)
			{
				CView* pView = pDocument->GetNextView(pos);
				ASSERT_VALID(pView);
				RemoveDocument(pDocument);
				pView->GetParentFrame()->SendMessage(WM_CLOSE, 0, 0);
			}
		}
	}

	bool is_image_format = GFLBitmap::isImageFormat(lpszPathName);

	CDocument* pDocument = 0;
	if (is_image_format)
	{
		if (m_pDocImageClass == NULL)
		{
			ASSERT(FALSE);
			return NULL;
		}
		pDocument = (CDocument*)m_pDocImageClass->CreateObject();
	}
	else
	{
		if (m_pDocClass == NULL)
		{
			TRACE0("Error: you must override CDocTemplate::CreateNewDocument.\n");
			ASSERT(FALSE);
			return NULL;
		}
		pDocument = (CDocument*)m_pDocClass->CreateObject();
	}
	if (pDocument == NULL)
	{
		TRACE1("Warning: Dynamic create of document type %hs failed.\n",
			m_pDocClass->m_lpszClassName);
		return NULL;
	}
	ASSERT_KINDOF(CDocument, pDocument);
	AddDocument(pDocument);
	if (pDocument == NULL)
	{
		TRACE0("CDocTemplate::CreateNewDocument returned NULL.\n");
		AfxMessageBox(AFX_IDP_FAILED_TO_CREATE_DOC);
		return NULL;
	}
	ASSERT_VALID(pDocument);

	BOOL bAutoDelete = pDocument->m_bAutoDelete;
	pDocument->m_bAutoDelete = FALSE;   // don't destroy if something goes wrong
	CFrameWnd* pFrame = 0;
	if (is_image_format)
		pFrame = CreateNewImageFrame(pDocument, NULL);
	else
		pFrame = CreateNewFrame(pDocument, NULL);
	pDocument->m_bAutoDelete = bAutoDelete;
	if (pFrame == NULL)
	{
		AfxMessageBox(AFX_IDP_FAILED_TO_CREATE_DOC);
		delete pDocument;       // explicit delete on error
		return NULL;
	}
	ASSERT_VALID(pFrame);

	if (lpszPathName == NULL)
	{
		// create a new document - with default document name
		SetDefaultTitle(pDocument);

		// avoid creating temporary compound file when starting up invisible
		if (!bMakeVisible)
			pDocument->m_bEmbedded = TRUE;

		if (!pDocument->OnNewDocument())
		{
			// user has be alerted to what failed in OnNewDocument
			TRACE0("CDocument::OnNewDocument returned FALSE.\n");
			pFrame->DestroyWindow();
			return NULL;
		}

		// it worked, now bump untitled count
		m_nUntitledCount++;
	}
	else
	{
		// open an existing document
		CWaitCursor wait;
		if (!pDocument->OnOpenDocument(lpszPathName))
		{
			// user has be alerted to what failed in OnOpenDocument
			TRACE0("CDocument::OnOpenDocument returned FALSE.\n");
			pFrame->DestroyWindow();
			return NULL;
		}
		pDocument->SetPathName(lpszPathName);
	}

	InitialUpdateFrame(pFrame, pDocument, bMakeVisible);
	return pDocument;
}

CFrameWnd* CSampleMultiDocTemplate::CreateNewImageFrame(CDocument* pDoc, CFrameWnd* pOther)
{
	ASSERT(m_nIDResource != 0); // must have a resource ID to load from
	CCreateContext context;
	context.m_pCurrentFrame = NULL;
	context.m_pCurrentDoc = pDoc;
	context.m_pNewViewClass = m_pViewImageClass;
	context.m_pNewDocTemplate = this;

	if (m_pFrameClass == NULL)
	{
		TRACE0("Error: you must override CDocTemplate::CreateNewFrame.\n");
		ASSERT(FALSE);
		return NULL;
	}
	CFrameWnd* pFrame = (CFrameWnd*)m_pFrameClass->CreateObject();
	if (pFrame == NULL)
	{
		TRACE1("Warning: Dynamic create of frame %hs failed.\n",
			m_pFrameClass->m_lpszClassName);
		return NULL;
	}
	ASSERT_KINDOF(CFrameWnd, pFrame);

	if (context.m_pNewViewClass == NULL)
		TRACE0("Warning: creating frame with no default view.\n");

	// create new from resource
	if (!pFrame->LoadFrame(m_nIDResource,
			WS_OVERLAPPEDWINDOW | FWS_ADDTOTITLE,   // default frame styles
			NULL, &context))
	{
		TRACE0("Warning: CDocTemplate couldn't create a frame.\n");
		// frame will be deleted in PostNcDestroy cleanup
		return NULL;
	}

	// it worked !
	return pFrame;
}

