#ifndef _INCLUDED_SCANNER_H
#define _INCLUDED_SCANNER_H

class TextFileBuffer;

class AbstractScanner
{
public:
	virtual void addKeyword(Ident keyword) = 0;
	virtual bool isTerminal(Ident name) = 0;

	virtual void initScanning() = 0;
	virtual void skipSpace(TextFileBuffer& text) = 0;
	virtual bool acceptEOF(TextFileBuffer& text) = 0;
	virtual bool acceptLiteral(TextFileBuffer& text, const char*) = 0;
	virtual bool acceptTerminal(TextFileBuffer& text, Ident name, AbstractParseTree& result) = 0;
};

class BasicScanner : public AbstractScanner
{
public:
	BasicScanner();

	virtual void addKeyword(Ident keyword);
	virtual bool isTerminal(Ident name);
	
	virtual void initScanning();
	virtual void skipSpace(TextFileBuffer& text);
	virtual bool acceptEOF(TextFileBuffer& text);
	virtual bool acceptLiteral(TextFileBuffer& text, const char* sym);
	virtual bool acceptTerminal(TextFileBuffer& text, Ident name, AbstractParseTree& result);

private:
	bool is_keyword(Ident ident);

	bool accept_ident(TextFileBuffer& text, Ident& ident, bool& is_keyword, const char* keyword);
	bool accept_ident(TextFileBuffer& text, AbstractParseTree& result);
	bool accept_string(TextFileBuffer& text, AbstractParseTree& result, bool c_string);
	bool accept_char(TextFileBuffer& text, AbstractParseTree& result);
	bool accept_int(TextFileBuffer& text, AbstractParseTree& result);
	bool accept_double(TextFileBuffer& text, AbstractParseTree& result);

	Ident _keywords[500];
	int _nr_keywords;

	TextFilePos _last_space_pos;
	TextFilePos _last_space_end_pos;
	TextFilePos _last_ident_pos;
	TextFilePos _last_ident_end_pos;
	bool _last_ident_is_keyword;
	Ident _last_ident;

	static Ident id_ident;
	static Ident id_string;
	static Ident id_cstring;
	static Ident id_char;
	static Ident id_int;
	static Ident id_double;
};

#endif // _INCLUDED_SCANNER_H
