#include "stdafx.h"
#include "gfl.h"
#include <math.h>


static int roundToInt(double v) { return (int)floor(v + 0.5); }


static void (GFLAPI *gfl_FreeBitmap)( GFL_BITMAP* bitmap ) = 0; 
static void (GFLAPI *gfl_GetDefaultLoadParams)( GFL_LOAD_PARAMS* params ) = 0; 
static GFL_ERROR (GFLAPI *gfl_LoadBitmap)( const char* filename, GFL_BITMAP** bitmap, const GFL_LOAD_PARAMS* params, GFL_FILE_INFORMATION* info ) = 0;
static void (GFLAPI *gfl_GetDefaultSaveParams)( GFL_SAVE_PARAMS* params ) = 0; 
static GFL_INT32 (GFLAPI *gfl_GetFormatIndexByName)( const char* name ) = 0;
static GFL_ERROR (GFLAPI *gfl_SaveBitmap)( char* filename, const GFL_BITMAP *bitmap, const GFL_SAVE_PARAMS* params ) = 0; 
static GFL_BITMAP* (GFLAPI *gfl_AllockBitmap)( GFL_BITMAP_TYPE type, GFL_INT32 width, GFL_INT32 height, GFL_UINT32 line_padding, const GFL_COLOR*  color ) = 0; 
static GFL_ERROR (GFLAPI *gfl_GetColorAt)( const GFL_BITMAP* src, GFL_INT32 x, GFL_INT32 y, GFL_COLOR* color ) = 0;
static GFL_ERROR (GFLAPI *gfl_SetColorAt)( GFL_BITMAP* src, GFL_INT32 x, GFL_INT32 y, const GFL_COLOR* color ) = 0;
static bool gfl_initialized = false;

static int nr_formats = 0;
static GFL_FORMAT_INFORMATION *format_information = 0;

bool GFLBitmap::initialized()
{
	return gfl_initialized;
}

void GFLBitmap::clear()
{
	if (m_gflBitmap != 0)
		gfl_FreeBitmap(m_gflBitmap);
	m_gflBitmap = 0;
	if (m_dib_info != 0)
		free(m_dib_info);
	m_dib_info = 0;
}

int GFLBitmap::width()
{
	return m_gflBitmap ? m_gflBitmap->Width : 0;
}

int GFLBitmap::height()
{
	return m_gflBitmap ? m_gflBitmap->Height : 0;
}


bool GFLBitmap::load(const char* filename)
{
	if (!gfl_initialized)
		return false;

	GFL_LOAD_PARAMS load_params; 

	clear();
	
	gfl_GetDefaultLoadParams( &load_params ); 
	//load_params.Flags |= GFL_LOAD_SKIP_ALPHA; 
	load_params.Origin = GFL_BOTTOM_LEFT; 
	load_params.ColorModel = GFL_BGR; 
	//load_params.LinePadding = 4; 
	return gfl_LoadBitmap( filename, &m_gflBitmap, &load_params, NULL ) == GFL_NO_ERROR;
}

bool GFLBitmap::save(const char* filename)
{
	if (m_gflBitmap == 0)
		return false;

	GFL_SAVE_PARAMS save_params;

	gfl_GetDefaultSaveParams( &save_params );
	const char *extension = "png";
	for (const char *s = filename; *s != '\0'; s++)
		if (*s == '.')
			extension = s+1;
	save_params.FormatIndex = gfl_GetFormatIndexByName(extension);
	return gfl_SaveBitmap( (char*)filename, m_gflBitmap, &save_params ) == GFL_NO_ERROR;
}

void GFLBitmap::newImage(int width, int height)
{
	if (!gfl_initialized)
		return;

	clear();

	m_gflBitmap = gfl_AllockBitmap(GFL_BGR, width, height, 4, 0);
	m_gflBitmap->Origin = GFL_BOTTOM_LEFT;
}

void GFLBitmap::drawImageStretch(HDC hdc, int XDest, int YDest, int nDestWidth, int nDestHeight,
							     int XSrc, int YSrc)
{
	if (m_gflBitmap == 0)
		return;

	// Calculate DIB info
	if (m_dib_info == 0)
	{
		m_dib_info = (BITMAPINFOHEADER*)calloc( 1, sizeof(BITMAPINFOHEADER) + 256 * sizeof(RGBQUAD) );
		if (m_dib_info == 0)
			return;
		RGBQUAD *colormap = (RGBQUAD *)m_dib_info + sizeof(BITMAPINFOHEADER);
		int bytes_per_line = 0;
		
		m_dib_info->biSize   = sizeof(BITMAPINFOHEADER); 
		m_dib_info->biWidth  = m_gflBitmap->Width; 
		m_dib_info->biHeight = m_gflBitmap->Height; 
		m_dib_info->biPlanes = 1; 
		
		if ( m_gflBitmap->Type == GFL_BINARY )
		{
			bytes_per_line = ((m_gflBitmap->Width + 31) / 32) * 4; 
			m_dib_info->biClrUsed = 2;
			m_dib_info->biBitCount = 1; 
		}
		else
		if ( m_gflBitmap->Type == GFL_GREY
			|| m_gflBitmap->Type == GFL_COLORS )
		{
			bytes_per_line = (m_gflBitmap->Width + 3) & ~3; 
			m_dib_info->biClrUsed = 256;
			m_dib_info->biBitCount = 8; 
		}
		else
		{
			bytes_per_line = (m_gflBitmap->Width * 3 + 3) & ~3; 
			m_dib_info->biClrUsed = 0;
			m_dib_info->biBitCount = 24; 
		}
		m_dib_info->biCompression = BI_RGB; 
		m_dib_info->biSizeImage = bytes_per_line * m_gflBitmap->Height; 
		m_dib_info->biClrImportant = 0; 
			
		if ( m_dib_info->biClrUsed > 0 )
		{
			if ( m_gflBitmap->Type == GFL_BINARY )
			{
				colormap[0].rgbRed   = 0; 
				colormap[0].rgbGreen = 0; 
				colormap[0].rgbBlue  = 0;

				colormap[1].rgbRed   = 255; 
				colormap[1].rgbGreen = 255; 
				colormap[1].rgbBlue  = 255;
			}
			else
			if ( m_gflBitmap->Type == GFL_GREY )
			{
				for ( int i=0; i<m_gflBitmap->ColorUsed; i++ )
				{
					colormap[i].rgbRed   = (i * 255) / (m_gflBitmap->ColorUsed-1); 
					colormap[i].rgbGreen = colormap[i].rgbRed; 
					colormap[i].rgbBlue  = colormap[i].rgbRed; 
				}
			}
			else
			{
				for ( int i=0; i<m_gflBitmap->ColorUsed; i++ )
				{
					colormap[i].rgbRed   = m_gflBitmap->ColorMap->Red[i]; 
					colormap[i].rgbGreen = m_gflBitmap->ColorMap->Green[i]; 
					colormap[i].rgbBlue  = m_gflBitmap->ColorMap->Blue[i]; 
				}
			}
		}
	}

	StretchDIBits( hdc, XDest, YDest, nDestWidth, nDestHeight,
				   XSrc, YSrc, m_gflBitmap->Width, m_gflBitmap->Height, 
				   m_gflBitmap->Data, (BITMAPINFO *)m_dib_info, DIB_RGB_COLORS, SRCCOPY ); 
}

void GFLBitmap::getInterpolatedColor(double x, double y, double &r, double &g, double &b)
{
	r = g = b = 0.0;

	if (m_gflBitmap == 0)
		return;

	GFL_COLOR color;
	int x_int = roundToInt(x);
	int y_int = roundToInt(y);
	double x_frac = x - x_int;
	double y_frac = y - y_int;
	double weight = 0;
	if (gfl_GetColorAt(m_gflBitmap, x_int, y_int, &color) == GFL_NO_ERROR)
	{
		double w = (1-x_frac)*(1-y_frac);
		weight += w;
		r += w * color.Red;
		g += w * color.Green;
		b += w *color.Blue;
	}
	if (gfl_GetColorAt(m_gflBitmap, x_int+1, y_int, &color) == GFL_NO_ERROR)
	{
		double w = x_frac*(1-y_frac);
		weight += w;
		r += w * color.Red;
		g += w * color.Green;
		b += w *color.Blue;
	}
	if (gfl_GetColorAt(m_gflBitmap, x_int, y_int+1, &color) == GFL_NO_ERROR)
	{
		double w = (1-x_frac)*y_frac;
		weight += w;
		r += w * color.Red;
		g += w * color.Green;
		b += w *color.Blue;
	}
	if (gfl_GetColorAt(m_gflBitmap, x_int+1, y_int+1, &color) == GFL_NO_ERROR)
	{
		double w = x_frac*y_frac;
		weight += w;
		r += w * color.Red;
		g += w * color.Green;
		b += w *color.Blue;
	}
	if (weight > 0.0 && weight <= 0.9999)
	{
		r /= weight;
		g /= weight;
		b /= weight;
	}
}

void GFLBitmap::setColor(int x, int y, int r, int g, int b)
{
	if (m_gflBitmap == 0)
		return;

	GFL_COLOR color;
	color.Red = r;
	color.Green = g;
	color.Blue = b;
	gfl_SetColorAt(m_gflBitmap, x, y, &color);
}


void GFLBitmap::init()
{
#define GETPROCADDRESS(V,R,P,N) V = (R (GFLAPI *)P)GetProcAddress(hModule, N); if (V == 0){ TRACE("%s not found\n", N); ASSERT(0); return; }
#define GETPROCADDRESSLOCAL(V,R,P,N) R (GFLAPI *V)P = (R (GFLAPI *)P)GetProcAddress(hModule, N); if (V == 0){ TRACE("%s not found\n", N); ASSERT(0); return; }

	HMODULE hModule = ::LoadLibrary(_T("libgfl311.dll"));
	if (hModule == 0)
	{
		TRACE("libgfl311.dll not found\n");
		return;
	}
	GETPROCADDRESSLOCAL(gfl_LibraryInit, GFL_ERROR, (void), "gflLibraryInit");

	if (gfl_LibraryInit() != GFL_NO_ERROR)
	{
		TRACE("LibraryInit failed\n");
		return;
	}

	GETPROCADDRESSLOCAL(gfl_GetNumberOfFormat, GFL_INT32, (void), "gflGetNumberOfFormat");
	GETPROCADDRESSLOCAL(gfl_GetFormatInformationByIndex, GFL_ERROR, (GFL_INT32 index, GFL_FORMAT_INFORMATION* info), "gflGetFormatInformationByIndex");

	GETPROCADDRESS(gfl_FreeBitmap, void, ( GFL_BITMAP* bitmap ), "gflFreeBitmap"); 
	GETPROCADDRESS(gfl_GetDefaultLoadParams, void, ( GFL_LOAD_PARAMS* params ), "gflGetDefaultLoadParams"); 
	GETPROCADDRESS(gfl_LoadBitmap, GFL_ERROR, ( const char* filename, GFL_BITMAP** bitmap, const GFL_LOAD_PARAMS* params, GFL_FILE_INFORMATION* info ), "gflLoadBitmap"); 
	GETPROCADDRESS(gfl_GetDefaultSaveParams, void, ( GFL_SAVE_PARAMS* params ), "gflGetDefaultSaveParams"); 
	GETPROCADDRESS(gfl_GetFormatIndexByName, GFL_INT32, ( const char* name ), "gflGetFormatIndexByName");
	GETPROCADDRESS(gfl_SaveBitmap, GFL_ERROR, ( char* filename, const GFL_BITMAP *bitmap, const GFL_SAVE_PARAMS* params ), "gflSaveBitmap"); 
	GETPROCADDRESS(gfl_AllockBitmap, GFL_BITMAP*, ( GFL_BITMAP_TYPE type, GFL_INT32 width, GFL_INT32 height, GFL_UINT32 line_padding, const GFL_COLOR*  color ), "gflAllockBitmap"); 
	GETPROCADDRESS(gfl_GetColorAt, GFL_ERROR, ( const GFL_BITMAP* src, GFL_INT32 x, GFL_INT32 y, GFL_COLOR* color ), "gflGetColorAt");
	GETPROCADDRESS(gfl_SetColorAt, GFL_ERROR, ( GFL_BITMAP* src, GFL_INT32 x, GFL_INT32 y, const GFL_COLOR* color ), "gflSetColorAt");

	// GFL_BITMAP* (GFLAPI *gfl_AllockBitmap)

	nr_formats = gfl_GetNumberOfFormat();
	if (nr_formats > 0)
	{
		format_information = new GFL_FORMAT_INFORMATION[nr_formats];

	    for (int i = 0; i < nr_formats; i++)
			gfl_GetFormatInformationByIndex(i, &format_information[i]);
	}

	gfl_initialized = true;
}

bool GFLBitmap::isImageFormat(LPCTSTR lpszPathName)
{
	if (lpszPathName == 0)
		return false;
	if (format_information == 0)
		return false;

	LPCTSTR ext = 0;
	for (LPCTSTR s = lpszPathName; *s != _T('\0'); s++)
		if (*s == _T('.'))
			ext = s + 1;
	if (ext == 0 || *ext == _T('\0'))
		return false;

    for (int i = 0; i < nr_formats; i++)
    {
		if (!(format_information[i].Status & GFL_READ ))
		    continue;
      
		for (int k = 0; k < format_information[i].NumberOfExtension; k++)
			if (stricmp(ext, format_information[i].Extension[k]) == 0)
				return true;
    }
	return false;
}
